@@ -25,6 +25,7 @@
 #include <gtk/gtksignal.h>
 
 
+#define MIN_SPIN_BUTTON_WIDTH           30
 #define ARROW_SIZE                      11
 #define SPIN_BUTTON_INITIAL_TIMER_DELAY 100
 #define SPIN_BUTTON_TIMER_DELAY         20
@@ -59,6 +60,8 @@ static gint gtk_spin_button_enter_notify   (GtkWidget          *widget,
 					    GdkEventCrossing   *event);
 static gint gtk_spin_button_leave_notify   (GtkWidget          *widget,
 					    GdkEventCrossing   *event);
+static gint gtk_spin_button_focus_out      (GtkWidget          *widget,
+					    GdkEventFocus      *event);
 static void gtk_spin_button_draw_arrow     (GtkSpinButton      *spin_button, 
 					    guint               arrow);
 static gint gtk_spin_button_timer          (GtkSpinButton      *spin_button);
@@ -70,6 +73,8 @@ static void gtk_spin_button_value_changed  (GtkWidget          *widget,
 static gint gtk_spin_button_key_press      (GtkWidget          *widget,
 					    GdkEventKey        *event);
 static void gtk_spin_button_update         (GtkSpinButton      *spin_button);
+static void gtk_spin_button_changed        (GtkEntry           *entry);
+static void gtk_spin_button_activate       (GtkEntry           *entry);
 
 
 static GtkWidgetClass *parent_class = NULL;
@@ -96,7 +101,6 @@ gtk_spin_button_get_type ()
       spin_button_type = gtk_type_unique (gtk_entry_get_type (), 
 					  &spin_button_info);
     }
-
   return spin_button_type;
 }
 
@@ -105,9 +109,11 @@ gtk_spin_button_class_init (GtkSpinButtonClass *class)
 {
   GtkObjectClass *object_class;
   GtkWidgetClass *widget_class;
+  GtkEntryClass  *entry_class;
 
   object_class = (GtkObjectClass*) class;
   widget_class = (GtkWidgetClass*) class;
+  entry_class  = (GtkEntryClass*)  class; 
 
   parent_class = gtk_type_class (gtk_entry_get_type ());
 
@@ -127,6 +133,10 @@ gtk_spin_button_class_init (GtkSpinButtonClass *class)
   widget_class->key_press_event = gtk_spin_button_key_press;
   widget_class->enter_notify_event = gtk_spin_button_enter_notify;
   widget_class->leave_notify_event = gtk_spin_button_leave_notify;
+  widget_class->focus_out_event = gtk_spin_button_focus_out;
+
+  entry_class->changed = gtk_spin_button_changed;
+  entry_class->activate = gtk_spin_button_activate;
 }
 
 static void
@@ -136,11 +146,12 @@ gtk_spin_button_init (GtkSpinButton *spin_button)
   spin_button->panel      = NULL;
 
   spin_button->timer = 0;
-  spin_button->climb_rate = 0;
+  spin_button->climb_rate = 0.0;
   spin_button->timer_step = 0.0;
   
-  spin_button->update_policy = GTK_UPDATE_ALWAYS;
-  spin_button->update_policy = GTK_UPDATE_IF_VALID;
+  spin_button->update_policy = GTK_UPDATE_ALWAYS | GTK_UPDATE_SNAP_TO_TICKS;
+
+  spin_button->snapped = 0;
   spin_button->in_child = 2;
   spin_button->click_child = 2;
   spin_button->button = 0;
@@ -162,8 +173,7 @@ gtk_spin_button_new (GtkAdjustment *adjustment,
   spin = gtk_type_new (gtk_spin_button_get_type ());
 
   if (!adjustment)
-    adjustment = (GtkAdjustment*) gtk_adjustment_new (0.0, 0.0, 0.0, 0.0,
-						      0.0, 0.0);
+    adjustment = (GtkAdjustment*) gtk_adjustment_new (0, 0, 0, 0, 0, 0);
 
   gtk_spin_button_set_adjustment (GTK_SPIN_BUTTON (spin), adjustment);
   spin->digits = digits;
@@ -177,14 +187,10 @@ gtk_spin_button_new (GtkAdjustment *adjustment,
 static void
 gtk_spin_button_destroy (GtkObject *object)
 {
-  GtkSpinButton *spin;
-  
   g_return_if_fail (object != NULL);
   g_return_if_fail (GTK_IS_SPIN_BUTTON (object));
 
-  spin = GTK_SPIN_BUTTON (object);
-
-  gtk_object_unref (GTK_OBJECT (spin->adjustment));
+  gtk_object_unref (GTK_OBJECT (GTK_SPIN_BUTTON (object)->adjustment));
   
   (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
 }
@@ -192,16 +198,12 @@ gtk_spin_button_destroy (GtkObject *object)
 static void
 gtk_spin_button_map (GtkWidget *widget)
 {
-  GtkSpinButton *spin;
-  
   g_return_if_fail (widget != NULL);
   g_return_if_fail (GTK_IS_SPIN_BUTTON (widget));
 
-  spin = GTK_SPIN_BUTTON (widget);
-  
   if (!GTK_WIDGET_MAPPED (widget))
     {
-      gdk_window_show (spin->panel);
+      gdk_window_show (GTK_SPIN_BUTTON (widget)->panel);
       GTK_WIDGET_CLASS (parent_class)->map (widget);
     }
 }
@@ -209,16 +211,12 @@ gtk_spin_button_map (GtkWidget *widget)
 static void
 gtk_spin_button_unmap (GtkWidget *widget)
 {
-  GtkSpinButton *spin;
-
   g_return_if_fail (widget != NULL);
   g_return_if_fail (GTK_IS_SPIN_BUTTON (widget));
 
-  spin = GTK_SPIN_BUTTON (widget);
-
   if (GTK_WIDGET_MAPPED (widget))
     {
-      gdk_window_hide (spin->panel);
+      gdk_window_hide (GTK_SPIN_BUTTON (widget)->panel);
       GTK_WIDGET_CLASS (parent_class)->unmap (widget);
     }
 }
@@ -292,22 +290,20 @@ gtk_spin_button_size_request (GtkWidget      *widget,
 
   GTK_WIDGET_CLASS (parent_class)->size_request (widget, requisition);
   
-  requisition->width += ARROW_SIZE + 2 * widget->style->klass->xthickness;
+  requisition->width = MIN_SPIN_BUTTON_WIDTH + ARROW_SIZE 
+    + 2 * widget->style->klass->xthickness;
 }
 
 static void
 gtk_spin_button_size_allocate (GtkWidget     *widget,
 			       GtkAllocation *allocation)
 {
-  GtkSpinButton *spin;
   GtkAllocation child_allocation;
 
   g_return_if_fail (widget != NULL);
   g_return_if_fail (GTK_IS_SPIN_BUTTON (widget));
   g_return_if_fail (allocation != NULL);
 
-  spin = GTK_SPIN_BUTTON (widget);
-
   child_allocation = *allocation;
   child_allocation.width = allocation->width - ARROW_SIZE  
     - 2 * widget->style->klass->xthickness;
@@ -324,8 +320,10 @@ gtk_spin_button_size_allocate (GtkWidget     *widget,
   child_allocation.y = allocation->y + 
     (allocation->height - widget->requisition.height) / 2;
 
-  gdk_window_move_resize (spin->panel, child_allocation.x,
-			  child_allocation.y, child_allocation.width, 
+  gdk_window_move_resize (GTK_SPIN_BUTTON (widget)->panel, 
+			  child_allocation.x,
+			  child_allocation.y,
+			  child_allocation.width,
 			  child_allocation.height); 
 }
 
@@ -342,7 +340,6 @@ gtk_spin_button_paint (GtkWidget    *widget,
 
   if (GTK_WIDGET_DRAWABLE (widget))
     {
-      
       gtk_draw_shadow (widget->style, spin->panel,
 		       GTK_STATE_NORMAL, GTK_SHADOW_OUT,
 		       0, 0, ARROW_SIZE + 2 * widget->style->klass->xthickness,
@@ -359,35 +356,26 @@ static void
 gtk_spin_button_draw (GtkWidget    *widget,
 		      GdkRectangle *area)
 {
-  GtkSpinButton *spin;
-
   g_return_if_fail (widget != NULL);
   g_return_if_fail (GTK_IS_SPIN_BUTTON (widget));
   g_return_if_fail (area != NULL);
 
 
   if (GTK_WIDGET_DRAWABLE (widget))
-    {
-      spin = GTK_SPIN_BUTTON (widget);
-      gtk_spin_button_paint (widget, area);
-    }
+    gtk_spin_button_paint (widget, area);
 }
 
 static gint
 gtk_spin_button_expose (GtkWidget      *widget,
 			GdkEventExpose *event)
 {
-  GtkSpinButton *spin;
-
   g_return_val_if_fail (widget != NULL, FALSE);
   g_return_val_if_fail (GTK_IS_SPIN_BUTTON (widget), FALSE);
   g_return_val_if_fail (event != NULL, FALSE);
 
   if (GTK_WIDGET_DRAWABLE (widget))
-    {
-      spin = GTK_SPIN_BUTTON (widget);
-      gtk_spin_button_paint (widget, &event->area);
-    }
+    gtk_spin_button_paint (widget, &event->area);
+
   return FALSE;
 }
 
@@ -402,7 +390,7 @@ gtk_spin_button_draw_arrow (GtkSpinButton *spin_button,
   g_return_if_fail (spin_button != NULL);
   g_return_if_fail (GTK_IS_SPIN_BUTTON (spin_button));
   
-  widget = GTK_WIDGET(spin_button);
+  widget = GTK_WIDGET (spin_button);
 
   if (GTK_WIDGET_DRAWABLE (spin_button))
     {
@@ -466,14 +454,12 @@ gtk_spin_button_enter_notify (GtkWidget        *widget,
       if (y <= widget->requisition.height / 2)
 	{
 	  spin->in_child = GTK_ARROW_UP;
-	  
 	  if (spin->click_child == 2) 
 	    gtk_spin_button_draw_arrow (spin, GTK_ARROW_UP);
 	}
       else
 	{
 	  spin->in_child = GTK_ARROW_DOWN;
-	  
 	  if (spin->click_child == 2) 
 	    gtk_spin_button_draw_arrow (spin, GTK_ARROW_DOWN);
 	}
@@ -509,6 +495,25 @@ gtk_spin_button_leave_notify (GtkWidget        *widget,
   return FALSE;
 }
 
+static gint
+gtk_spin_button_focus_out (GtkWidget     *widget,
+			   GdkEventFocus *event)
+{
+  GtkSpinButton *spin;
+
+  g_return_val_if_fail (widget != NULL, FALSE);
+  g_return_val_if_fail (GTK_IS_SPIN_BUTTON (widget), FALSE);
+  g_return_val_if_fail (event != NULL, FALSE);
+
+  spin = GTK_SPIN_BUTTON (widget);
+  if ((spin->update_policy & GTK_UPDATE_SNAP_TO_TICKS) && !spin->snapped)
+    gtk_spin_button_update (spin);
+
+  GTK_WIDGET_CLASS (parent_class)->focus_out_event (widget, event);
+  
+  return FALSE;
+}
+
 static gint
 gtk_spin_button_button_press (GtkWidget      *widget,
 			      GdkEventButton *event)
@@ -562,7 +567,6 @@ gtk_spin_button_button_press (GtkWidget      *widget,
 			 (GtkFunction) gtk_spin_button_timer, (gpointer) spin);
 		    }
 		}
-
 	      gtk_spin_button_draw_arrow (spin, GTK_ARROW_UP);
 	    }
 	  else 
@@ -594,14 +598,11 @@ gtk_spin_button_button_press (GtkWidget      *widget,
 			 (GtkFunction) gtk_spin_button_timer, (gpointer) spin);
 		    }
 		}
-
 	      gtk_spin_button_draw_arrow (spin, GTK_ARROW_DOWN);
 	    }
 	}
       else 
-	{
-	  GTK_WIDGET_CLASS (parent_class)->button_press_event (widget, event);
-	}
+	GTK_WIDGET_CLASS (parent_class)->button_press_event (widget, event);
     }
   return FALSE;
 }
@@ -732,7 +733,7 @@ gtk_spin_button_timer (GtkSpinButton *spin_button)
 	   (gpointer) spin_button);
 	return FALSE;
       }
-      else if (spin_button->climb_rate > 0 && 
+      else if (spin_button->climb_rate > 0.0 && 
 	   spin_button->timer_step < spin_button->adjustment->page_increment)
 	{
 	  if (spin_button->timer_calls < MAX_TIMER_CALLS)
@@ -753,7 +754,7 @@ gtk_spin_button_spin (GtkSpinButton *spin_button,
 		      guint          direction,
 		      gfloat         step)
 {
-  gfloat new_value = 0;
+  gfloat new_value = 0.0;
 
   g_return_if_fail (spin_button != NULL);
   g_return_if_fail (GTK_IS_SPIN_BUTTON (spin_button));
@@ -846,15 +847,10 @@ gtk_spin_button_key_press (GtkWidget     *widget,
 	gtk_spin_button_spin (spin, GTK_ARROW_DOWN,
 			      spin->adjustment->page_increment);
       return TRUE;
-    case GDK_Return:
-      if (GTK_ENTRY (spin)->editable) {
-	gtk_signal_emit_stop_by_name (GTK_OBJECT (widget), "key_press_event");
-	gtk_spin_button_update (spin);
-	return TRUE;
-      }
     default:
-      return GTK_WIDGET_CLASS (parent_class)->key_press_event (widget, event);
+      break;
     }
+  return GTK_WIDGET_CLASS (parent_class)->key_press_event (widget, event);
 }
 
 static void 
@@ -869,35 +865,76 @@ gtk_spin_button_update (GtkSpinButton *spin_button)
 
   val = strtod (GTK_ENTRY (spin_button)->text, &error);
   
-  if (spin_button->update_policy == GTK_UPDATE_ALWAYS)
+  if ((spin_button->update_policy & GTK_UPDATE_IF_VALID) && 
+      (*error ||
+       val < spin_button->adjustment->lower ||
+       val > spin_button->adjustment->upper))
+    {
+      gtk_signal_emit_by_name (GTK_OBJECT (spin_button->adjustment),
+			       "value_changed"); 
+      return;
+    }
+
+  if (spin_button->update_policy & GTK_UPDATE_SNAP_TO_TICKS)
+    {
+      gfloat inc;
+
+      inc = spin_button->adjustment->step_increment;
+      tmp = (val - spin_button->adjustment->lower) / inc;
+      if (tmp - floor (tmp) < ceil (tmp) - tmp)
+	val = spin_button->adjustment->lower + floor (tmp) * inc;
+      else
+	val = spin_button->adjustment->lower + ceil (tmp) * inc;
+      spin_button->snapped = 1;
+    }
+  if (spin_button->update_policy & GTK_UPDATE_ALWAYS)
     {
       if (val < spin_button->adjustment->lower)
 	val = spin_button->adjustment->lower;
       else if (val > spin_button->adjustment->upper)
 	val = spin_button->adjustment->upper;
     }
-  else if (*error ||
-	   val < spin_button->adjustment->lower ||
-	   val > spin_button->adjustment->upper)
-    {
-      gtk_signal_emit_by_name (GTK_OBJECT (spin_button->adjustment),
-			       "value_changed"); 
-      return;
-    }
-  tmp = (val - spin_button->adjustment->lower) / 
-    spin_button->adjustment->step_increment;
-  if (tmp - floor (tmp) < ceil (tmp) - tmp)
-    val = spin_button->adjustment->lower + 
-      floor (tmp) * spin_button->adjustment->step_increment;
-  else
-    val = spin_button->adjustment->lower + 
-      ceil (tmp) * spin_button->adjustment->step_increment;
-  
+
   spin_button->adjustment->value = val;
   gtk_signal_emit_by_name (GTK_OBJECT (spin_button->adjustment),
 			   "value_changed"); 
 }
 
+static void
+gtk_spin_button_changed (GtkEntry *entry)
+{
+  g_return_if_fail (entry != NULL);
+  g_return_if_fail (GTK_IS_ENTRY (entry));
+
+  GTK_ENTRY_CLASS (parent_class)->changed (entry);
+  if (GTK_WIDGET_VISIBLE (GTK_WIDGET (entry)))
+    {
+      GtkSpinButton *spin;
+      gfloat val;
+      gchar *error = NULL;
+
+      spin = GTK_SPIN_BUTTON (entry);
+      spin->snapped = 0;
+
+      val = strtod (entry->text, &error);
+      if (val < spin->adjustment->lower)
+	val = spin->adjustment->lower;
+      else if (val > spin->adjustment->upper)
+	val = spin->adjustment->upper;
+      spin->adjustment->value = val;
+    }
+}
+
+static void
+gtk_spin_button_activate (GtkEntry *entry)
+{
+  g_return_if_fail (entry != NULL);
+  g_return_if_fail (GTK_IS_ENTRY (entry));
+
+  if (entry->editable)
+    gtk_spin_button_update (GTK_SPIN_BUTTON (entry));
+}
+
 void
 gtk_spin_button_set_adjustment (GtkSpinButton *spin_button,
 				GtkAdjustment *adjustment)
@@ -945,7 +982,6 @@ gtk_spin_button_set_digits (GtkSpinButton *spin_button,
   if (spin_button->digits != digits)
     {
       spin_button->digits = digits;
-
       if (GTK_WIDGET_VISIBLE (spin_button) && GTK_WIDGET_MAPPED (spin_button))
         gtk_signal_emit_by_name (GTK_OBJECT (spin_button->adjustment),
 				 "value_changed"); 
@@ -986,7 +1022,6 @@ gtk_spin_button_set_value (GtkSpinButton *spin_button,
   if (spin_button->adjustment->value != value)
     {
       spin_button->adjustment->value = value;
-
       if (GTK_WIDGET_VISIBLE (spin_button) && GTK_WIDGET_MAPPED (spin_button))
         gtk_signal_emit_by_name (GTK_OBJECT (spin_button->adjustment),
 				 "value_changed"); 
